/*
 */
#include "globus_common.h"
#include "globus_i_thread.h"

typedef struct globus_i_thread_s
{
    int				id;
    void **			key_table_values;
    
    /* These are used only during thread startup */
    void *			context;
    globus_thread_func_t	user_func;
    void *			user_arg;
    
    /* This is only used when this structure is on the free list */
    struct globus_i_thread_s   *next_free;
} globus_i_thread_t;

#define GLOBUS_I_THREAD_SIZE() \
    (sizeof(globus_i_thread_t) + (key_table_size * sizeof(void *)))


/*
 * globus_i_thread_self()
 *
 * Set *Thread (globus_i_thread_t **) to the calling thread.
 */
#define globus_i_thread_self(Thread) \
    *(Thread) = ((globus_i_thread_t *) cthread_data(cthread_self()))


/*
 * globus_i_thread_id()
 *
 * Set *Thread_ID (int *) to be the thread id of the calling thread.
 */
#define globus_i_thread_id(Thread_ID) \
{ \
    globus_i_thread_t *my_thread; \
    globus_i_thread_self(&my_thread); \
    *(Thread_ID) = my_thread->id; \
}


#ifndef GLOBUS_L_THREAD_KEY_TABLE_SIZE
#define GLOBUS_L_THREAD_KEY_TABLE_SIZE 16
#endif

static globus_thread_key_destructor_func_t *key_table_destructor_funcs;

/*
 * Thread specific data key maintainnance stuff
 */
static int		key_table_size;
static int		next_key;
static globus_mutex_t	key_mutex;
static globus_bool_t    globus_l_thread_already_initialized=GLOBUS_FALSE;

/*
 * Free list of globus_i_thread_t structures
 */
static globus_i_thread_t *	free_list;
static globus_mutex_t	free_list_mutex;
static int		next_thread_id;


/*
 * For globus_thread_once()
 */
static globus_mutex_t	thread_once_mutex;


static globus_bool_t		using_idle_thread = GLOBUS_FALSE;
static cthread_t	idle_thread;
static globus_bool_t		idle_thread_done;
static void *		idle_thread_func(void *arg);
static globus_bool_t		preemptive_threads = GLOBUS_TRUE;

static void *		thread_startup(void * arg);
static void (*idle_func_save)(void);


/*
 * globus_i_thread_pre_activate()
 *
 */
int
globus_i_thread_pre_activate( void )
{
    return GLOBUS_SUCCESS;
} /* globus_i_thread_pre_activate() */


/*
 * globus_l_thread_activate()
 */
int
globus_l_thread_activate()
{
    globus_i_thread_t *thread;
    int i;
    
    if(globus_l_thread_already_initialized)
    {
	return GLOBUS_SUCCESS;
    }
    globus_l_thread_already_initialized = GLOBUS_TRUE;

    globus_mutex_init(&globus_libc_mutex, (globus_mutexattr_t *) NULL);

    /* Initialize thread specific keys */
    key_table_size = GLOBUS_L_THREAD_KEY_TABLE_SIZE;
    next_key = 0;
    globus_mutex_init(&key_mutex, (globus_mutexattr_t *) NULL);
    
    /* Initialize the free list */
    free_list = (globus_i_thread_t *) NULL;
    globus_mutex_init(&free_list_mutex, (globus_mutexattr_t *) NULL);
    next_thread_id = 1;

    /* Initialize the thread_once_mutex */
    globus_mutex_init(&thread_once_mutex, (globus_mutexattr_t *) NULL);

    /* Initialize the key_table_destructor_funcs */
    GlobusThreadMalloc(
	globus_l_thread_activate(),
	key_table_destructor_funcs,
	globus_thread_key_destructor_func_t *,
	(key_table_size*sizeof(globus_thread_key_destructor_func_t)));
    
    for (i = 0; i < key_table_size; i++)
    {
	key_table_destructor_funcs[i] = NULL;
    }
    
    /* Initialize the globus_i_thread_t structure for this initial thread */
    GlobusThreadMalloc(globus_l_thread_activate(),
		       thread,
		       globus_i_thread_t *,
		       GLOBUS_I_THREAD_SIZE() );
    thread->id = 0;
    thread->key_table_values
	= (void **) (((char *) thread) + sizeof(globus_i_thread_t));
    for (i = 0; i < key_table_size; i++)
    {
	thread->key_table_values[i] = (void *) NULL;
    }
    
    cthread_set_data(cthread_self(), thread);

    return GLOBUS_SUCCESS;
} /* globus_l_thread_activate() */


/*
 * globus_l_thread_deactivate()
 */
int
globus_l_thread_deactivate(void)
{
    return GLOBUS_SUCCESS;
} /* globus_l_thread_deactivate() */

/*
 * terminate_thread()
 *
 * Put the thread's data structure onto the free_list, and exit
 * from the thread.
 */
static void
terminate_thread(globus_i_thread_t *thread, void *status)
{
    int i;
    globus_thread_key_destructor_func_t func;
    void *value;
    
    /* Call the thread specific data destructors */
    for (i = 0; i < key_table_size; i++)
    {
	func = key_table_destructor_funcs[i];
	value = thread->key_table_values[i];
	if (func && value)
	{
	    (*func)(value);
	}
    }

#ifdef BUILD_PROFILE
    /*
    log_thread_destruction(thread->id);
    */
#endif /* BUILD_PROFILE */    

    /* Free up the thread storage */
    globus_mutex_lock(&free_list_mutex);
    thread->next_free = free_list;
    free_list = thread;
    globus_mutex_unlock(&free_list_mutex);

    /* Exit the thread */
    cthread_exit((void *) status);
} /* terminate_thread() */


/*
 * thread_startup()
 *
 * This is the entry point for a new cthread that is created
 * by globus_thread_create().  It does some bookkeeping and then
 * calls the user function.  If the user function returns, then
 * it terminates the thread.
 */
static void *
thread_startup(void * arg)
{
    globus_i_thread_t *thread = (globus_i_thread_t *) arg;
    int i;
    void *status;

    /* Initialize the thread specific key table */
    for (i = 0; i < key_table_size; i++)
    {
	thread->key_table_values[i] = (void *) NULL;
    }

    /* Setup pointers so we can find the thread data again */
    cthread_set_data(cthread_self(), thread);

#ifdef BUILD_PROFILE
    /*
    log_thread_creation(thread->id);
    */
#endif /* BUILD_PROFILE */    
    
    /* Call the user function */
    status = (*thread->user_func)(thread->user_arg);

    /* Terminate the thread */
    terminate_thread(thread, status);

    return (0);
} /* thread_startup() */


/*
 * get_new_thread()
 *
 * Return a globus_i_thread_t for a new thread.
 */
static globus_i_thread_t *
get_new_thread(void)
{
    globus_i_thread_t *thread;
    
    globus_mutex_lock(&free_list_mutex);
    if (free_list != (globus_i_thread_t *) NULL)
    {
	thread = free_list;
	free_list = free_list->next_free;
    }
    else
    {
	GlobusThreadMalloc(globus_thread_create(),
			   thread,
			   globus_i_thread_t *,
			   GLOBUS_I_THREAD_SIZE() );
    }
    thread->id = next_thread_id++;
    globus_mutex_unlock(&free_list_mutex);

    return (thread);
} /* get_new_thread() */


/*
 * globus_thread_create()
 */
int
globus_thread_create(globus_thread_t *user_thread,
		     globus_threadattr_t *attr,
		     globus_thread_func_t func,
		     void *user_arg)
{
    globus_i_thread_t *thread;
    cthread_t cthread_id;

    thread = get_new_thread();

    /* Initialize the thread data */
    thread->key_table_values
	= (void **) (((char *) thread) + sizeof(globus_i_thread_t));
    
    /* Initialize the thread data that needs to be passed to the new thread */
    thread->user_func = func;
    thread->user_arg = user_arg;
    
    cthread_id = cthread_fork(thread_startup, thread);
    cthread_detach(cthread_id);

    if (user_thread)
    {
	*user_thread = *cthread_id;
    }
    
    return (0);
    
} /* globus_thread_create() */


/*
 * globus_preemptive_threadss
 *
 * Return GLOBUS_TRUE (non-zero) if we are using preemptive threads.
 */
globus_bool_t
globus_preemptive_threadss(void)
{
    return (preemptive_threads);
} /* globus_preemptive_threadss() */


/*
 * globus_thread_exit()
 */
void
globus_thread_exit(void *status)
{
    globus_i_thread_t *thread;
    globus_i_thread_self(&thread);
    terminate_thread(thread, status);
} /* globus_thread_exit() */


/*
 * globus_thread_key_create()
 */
int
globus_thread_key_create(globus_thread_key_t *key,
			 globus_thread_key_destructor_func_t func)
{
    int rc;
    
    globus_mutex_lock(&key_mutex);
    if (next_key >= key_table_size)
    {
	/* Attempt to allocate a key when the key name space is exhausted */
	rc = EAGAIN;
    }
    else
    {
	*key = next_key++;
	key_table_destructor_funcs[*key] = func;
	rc = 0;
    }
    globus_mutex_unlock(&key_mutex);
    return(rc);
} /* globus_thread_key_create() */


/*
 * globus_thread_setspecific()
 */
int
globus_thread_setspecific(globus_thread_key_t key,
			  void *value)
{
    globus_i_thread_t *thread;

    if(key < 0 || key >= key_table_size)
    {
	return GLOBUS_FAILURE;
    }

    globus_i_thread_self(&thread);
    thread->key_table_values[key] = value;
    return 0;
} /* globus_thread_setspecific() */


/*
 * globus_thread_getspecific()
 */
void *
globus_thread_getspecific(globus_thread_key_t key)
{
    void *value;
    globus_i_thread_t *thread;
    
    if(key < 0 || key >= key_table_size)
    {
	return GLOBUS_NULL;
    }

    globus_i_thread_self(&thread);
    value = thread->key_table_values[key];
    return (value);
} /* globus_thread_getspecific() */


/*
 * globus_i_thread_actual_thread_once()
 */
int
globus_i_thread_actual_thread_once(globus_thread_once_t *once_control,
		       void (*init_routine)(void))
{
    int rc=-1;
    globus_mutex_lock(&thread_once_mutex);
    if (*once_control)
    {
	/* Someone beat us to it.  */
	rc = 0;
    }
    else
    {
	/* We're the first one here */
	(*init_routine)();
	*once_control = 1;
	rc = 0;
    }
    globus_mutex_unlock(&thread_once_mutex);
    return (rc);
} /* globus_i_thread_actual_thread_once() */



/*
 * Now provide the function versions of each of the macros
 * for when USE_MACROS is not defined.
 */
#undef globus_thread_yield
void
globus_thread_yield(void)
{
    globus_macro_thread_yield();
} /* globus_thread_yield() */

#undef globus_thread_self
globus_thread_t
globus_thread_self(void)
{
    return (globus_macro_thread_self());
}

#undef globus_thread_equal
int
globus_thread_equal(globus_thread_t t1,
			globus_thread_t t2)
{
    return (globus_macro_thread_equal(t1,t2));
}
		       
#undef globus_thread_once
int
globus_thread_once(globus_thread_once_t *once_control,
		       void (*init_routine)(void))
{
    return (globus_macro_thread_once(once_control, init_routine));
}

#undef globus_i_am_only_thread
globus_bool_t
globus_i_am_only_thread(void)
{
    /*
     * return FALSE always because we never know what the number of
     * running threads is, so we can never assume this is the only
     * thread.
     */
    return GLOBUS_FALSE;
}


#undef globus_mutex_init
int
globus_mutex_init(globus_mutex_t *mutex, globus_mutexattr_t *attr)
{
    int rc;
    rc = globus_macro_mutex_init(mutex, attr);
    return rc;
}
#undef globus_mutex_destroy
int
globus_mutex_destroy(globus_mutex_t *mutex)
{
    int rc;
    rc = globus_macro_mutex_destroy(mutex);
    return rc;
}

#undef globus_mutex_lock
int
globus_mutex_lock(globus_mutex_t *mutex)
{
    int rc;
    rc = globus_macro_mutex_lock(mutex);
    return rc;
}

#undef globus_mutex_trylock
int
globus_mutex_trylock(globus_mutex_t *mutex)
{
    int rc;
    rc = globus_macro_mutex_trylock(mutex);
    return rc;
}

#undef globus_mutex_unlock
int
globus_mutex_unlock(globus_mutex_t *mutex)
{
    int rc;
    rc = globus_macro_mutex_unlock(mutex);
    return rc;
}

#undef globus_cond_init
int
globus_cond_init(globus_cond_t *cond, globus_condattr_t *attr)
{
    int rc;
    
#ifdef TARGET_ARCH_NEXTSTEP
    globus_macro_cond_init(cond, attr);
    rc = 0;
#else    
    rc = globus_macro_cond_init(cond, attr);
#endif
    return (rc);
}

#undef globus_cond_destroy
int
globus_cond_destroy(globus_cond_t *cond)
{
    int rc;
#ifdef TARGET_ARCH_NEXTSTEP
    globus_macro_cond_destroy(cond);
    rc = 0;
#else
    rc = globus_macro_cond_destroy(cond);
#endif
    return (rc);
}

#undef globus_cond_signal
int
globus_cond_signal(globus_cond_t *cond)
{
    int rc;
#ifdef TARGET_ARCH_NEXTSTEP
    globus_macro_cond_signal(cond);
    rc = 0;
#else
    rc = globus_macro_cond_signal(cond);
#endif
    return (rc);
}

#undef globus_cond_broadcast
int
globus_cond_broadcast(globus_cond_t *cond)
{
    int rc;
#ifdef TARGET_ARCH_NEXTSTEP
    globus_macro_cond_broadcast(cond);
    rc = 0;
#else
    rc = globus_macro_cond_broadcast(cond);
#endif
    return (rc);
}

#undef globus_cond_wait
int
globus_cond_wait(globus_cond_t *cond,
		     globus_mutex_t *mutex)
{
    int rc;
    rc = globus_macro_cond_wait(cond, mutex);
    return (rc);
}

void
globus_thread_prefork(void)
{
/* Do Nothing */
}

void
globus_thread_postfork(void)
{
/* Do Nothing */
}
