/******************************************************************************
CIF_I_Comm_Connection_TCP.C

Description:

  CIF_Comm Connection class for TCP (reliable ordered unicast)

CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Comm-C++-Current/library/CIF_I_Comm_Connection_TCP.C,v $
  $Date: 1998/11/05 19:08:18 $
  $Revision: 1.8 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_I_Comm_Connection_TCP.H"

#include <string.h>


/*
 * CIF_I_Comm_Connection_TCP::establish_Connection()
 */
CIF_I_Comm_Connection::Connection_Flow_Mask
CIF_I_Comm_Connection_TCP::open_Connection(
    ARG_IN const char *				url,
    ARG_INMOD
    CIF_I_Comm_Endpoint::Properties &		ep_properties,
    ARG_IN const CIF_Toolkit_Timeout &		timeout)
{
    CIF_Bool					rc;

    /*
     * Parse URL, verifying that it is a valid URL for this protocol module
     */
    globus_url_t				url_info;

    if (globus_url_parse((char *) url, &url_info) != GLOBUS_SUCCESS)
    {
	CIF_Comm::status = CIF_Comm_Status::URL_FORMAT_BAD;
	return CIF_I_Comm_Connection::CONNECTION_UNESTABLISHED;
    }

    if (strcmp(url_info.scheme, CIF_COMM_PROTOCOL_TCP_STR) != 0)
    {
	globus_url_destroy(&url_info);
	
	CIF_Comm::status = CIF_Comm_Status::URL_SCHEME_UNSUPPORTED;
	return CONNECTION_UNESTABLISHED;
    }

    /*
     * Establish a connection with the listener
     */
    CIF_I_Comm_Startpoint			remote_ctrl_sp;
    CIF_I_Comm_Startpoint &			remote_comm_sp = l_sp;

    rc = remote_ctrl_sp.activate(url_info.host,
				 url_info.port);
    globus_url_destroy(&url_info);
    if (rc == false)
    {
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }
    
    /*
     * Establish a local endpoints for communicating control information and
     * data
     */
    CIF_I_Comm_Endpoint				local_ctrl_ep;
    CIF_I_Comm_Startpoint			local_ctrl_sp;
    CIF_I_Comm_Endpoint &			local_comm_ep = l_ep;
    CIF_I_Comm_Startpoint			local_comm_sp;
    
    if (local_ctrl_ep.activate(CIF_I_Comm_Endpoint::Properties()) == false)
    {
	remote_ctrl_sp.deactivate();
	
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }
    
    if (local_ctrl_sp.activate(local_ctrl_ep) == false)
    {
	local_ctrl_ep.deactivate();
	remote_ctrl_sp.deactivate();
	
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }

    if (local_comm_ep.activate(ep_properties) == false)
    {
	local_ctrl_sp.deactivate();
	local_ctrl_ep.deactivate();
	remote_ctrl_sp.deactivate();
	
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }

    if (local_comm_sp.activate(local_comm_ep) == false)
    {
	local_comm_ep.deactivate();
	local_ctrl_sp.deactivate();
	local_ctrl_ep.deactivate();
	remote_ctrl_sp.deactivate();

	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }
	
    /*
     * Send a startpoints bound to our endpoints
     */
    globus_nexus_buffer_t			send_buf;
    
    globus_nexus_buffer_init(&send_buf,
			     local_ctrl_sp.sizeof_Startpoint() +
			     local_comm_sp.sizeof_Startpoint(),
			     0);

    local_ctrl_sp.put_Startpoint(send_buf);
    local_comm_sp.put_Startpoint(send_buf);
    
    rc = remote_ctrl_sp.send(send_buf);
    remote_ctrl_sp.deactivate();
    if (rc == false)
    {
	local_comm_ep.deactivate();
	local_ctrl_ep.deactivate();

	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }

    /*
     * Wait for a startpoint from the remote process
     */
    CIF_Bool					status;
    globus_nexus_buffer_t			recv_buf;

    status = true;

    if (local_ctrl_ep.receive(recv_buf, timeout) == true)
    {
	if (remote_comm_sp.activate(recv_buf) == false)
	{
	    status = false;
	}
    }
    else
    {
	status = false;
    }
    
    globus_nexus_buffer_destroy(&recv_buf);
    
    if (!status)
    {
	CIF_Comm_Status				status;
	
	status = CIF_Comm::status;
	
	local_ctrl_ep.deactivate();
	local_comm_ep.deactivate();

	CIF_Comm::status = status;
	return CONNECTION_UNESTABLISHED;
    }

    CIF_Comm::status = CIF_Comm_Status::SUCCESS;
    return CONNECTION_BIDIRECTIONAL;
}
/* CIF_I_Comm_Connection::establish_Connection() */


/*
 * CIF_I_Comm_Connection_TCP::establish_Connection()
 */
CIF_I_Comm_Connection::Connection_Flow_Mask
CIF_I_Comm_Connection_TCP::open_Connection(
    ARG_INOUT globus_nexus_buffer_t &		nx_buf,
    ARG_INMOD
    CIF_I_Comm_Endpoint::Properties &		ep_properties,
    ARG_IN const CIF_Toolkit_Timeout &		timeout)
{
    CIF_Bool					rc;

    CIF_I_Comm_Startpoint			remote_ctrl_sp;
    CIF_I_Comm_Startpoint &			remote_comm_sp = l_sp;
    CIF_I_Comm_Endpoint &			local_comm_ep = l_ep;
    CIF_I_Comm_Startpoint			local_comm_sp;

    /*
     * Get remote startpoints
     */
    remote_ctrl_sp.activate(nx_buf);
    remote_comm_sp.activate(nx_buf);

    globus_nexus_buffer_destroy(&nx_buf);
    
    /*
     * Establish a local endpoint for communicating data
     */
    if (local_comm_ep.activate(ep_properties) == false)
    {
	remote_comm_sp.deactivate();
	remote_ctrl_sp.deactivate();
	
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }

    if (local_comm_sp.activate(local_comm_ep) == false)
    {
	local_comm_ep.deactivate();
	remote_comm_sp.deactivate();
	remote_ctrl_sp.deactivate();
	
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }
	
    /*
     * Send a startpoint bound to our endpoint
     */
    globus_nexus_buffer_t			send_buf;
    
    globus_nexus_buffer_init(&send_buf,
			     local_comm_sp.sizeof_Startpoint(),
			     0);

    local_comm_sp.put_Startpoint(send_buf);
    
    rc = remote_ctrl_sp.send(send_buf);
    remote_ctrl_sp.deactivate();
    if (rc == false)
    {
	remote_comm_sp.deactivate();
	local_comm_ep.deactivate();

	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return CONNECTION_UNESTABLISHED;
    }

    CIF_Comm::status = CIF_Comm_Status::SUCCESS;
    return CONNECTION_BIDIRECTIONAL;
}
/* CIF_I_Comm_Connection::establish_Connection() */
